use std::collections::BTreeMap;
use std::sync::Arc;

use clap::{Parser, Subcommand};
use serde::{Deserialize, Serialize};
use tokio::{
    io::{AsyncBufReadExt, AsyncWriteExt, BufReader, Result},
    net::{UnixListener, UnixStream},
    sync::{self, mpsc, RwLock},
};
#[allow(unused_imports)]
use tracing::{error, info, instrument, warn};

static LICENSE: &'static str = include_str!("../LICENSE");

const IPC_SOCKET_PATH: &str = "/tmp/swimd.sock";

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq, PartialOrd, Ord, Clone)]
enum SockType {
    Udp,
    Tcp,
}

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq, PartialOrd, Ord, Clone)]
struct PortMapping {
    sock_type: SockType,
    from: u32,
    to: u32,
}

#[derive(Debug, Serialize, Deserialize, Default, Clone)]
struct Config {
    mappings: BTreeMap<u32, PortMapping>,
}

#[derive(Debug)]
struct AppState {
    active_config: Arc<RwLock<Config>>,
    config: Arc<RwLock<Config>>,
    notify: sync::mpsc::Sender<()>, // Used to tell the ipc server to reload
}

#[derive(Debug, Serialize, Deserialize)]
enum ClientOp {
    Export,
    Import,
    Add(PortMapping),
    Del(Vec<u32>),
    Reload,
}

#[derive(Debug, Serialize, Deserialize)]
struct IPCRequest {
    op: ClientOp,
}

#[derive(Debug, Serialize, Deserialize)]
enum ClientResp<S: AsRef<str>> {
    Ok(Option<S>),
    Error(S),
}

impl<S: AsRef<str> + Serialize> ClientResp<S> {
    async fn write_to<W>(&self, mut writer: W) -> Result<()>
    where
        W: AsyncWriteExt + std::marker::Unpin,
    {
        writer
            .write_all(serde_json::to_string(self)?.as_bytes())
            .await
    }
}

#[instrument]
async fn handle_ipc_client(mut conn: UnixStream, app: Arc<AppState>) {
    let config = &app.config;
    let (reader, mut writer) = conn.split();
    let mut reader = BufReader::new(reader);
    let mut buf = String::new();

    if reader.read_line(&mut buf).await.unwrap() == 0 {
        return;
    }

    serde_json::to_string(&*config.read().await).unwrap();

    let resp: ClientResp<_> = match serde_json::from_str::<IPCRequest>(&buf) {
        Ok(req) => match req.op {
            ClientOp::Del(ni) => {
                let cfg = config.write().await;
                let km: Vec<String> = ni
                    .iter()
                    .filter(|i| cfg.mappings.contains_key(i))
                    .map(|n| n.to_string())
                    .collect();
                let kml = km.join(",");
                if !km.len() == 0 {
                    error!("Bad rule ids when deleting: {}", kml);

                    ClientResp::Error(format!("Some route IDs do not exist: {}", kml))
                } else {
                    for i in ni {
                        info!("Removed ruleset {i}!");
                    }
                    ClientResp::Ok(Some(format!("Removed routes: {kml}")))
                }
            }
            ClientOp::Export => {
                let cfg = serde_json::to_string(&*config.read().await).unwrap();
                if let Err(e) = writer.write_all(cfg.as_bytes()).await {
                    error!("Error writing to ipc socket! {e}");
                }
                ClientResp::Ok(None)
            }
            ClientOp::Import => todo!(),
            ClientOp::Add(port_mapping) => {
                let mut cfg = config.write().await;
                let next = cfg.mappings.keys().last().unwrap_or(&0) + 1;
                cfg.mappings.insert(next, port_mapping);
                ClientResp::Ok(None)
            }
            ClientOp::Reload => {
                app.notify.send(()).await.unwrap();
                ClientResp::Ok(None)
            }
        },
        Err(e) => {
            error!("Couldn't parse ipc request: {}", e);
            ClientResp::Error("Bad Request".to_string())
        }
    };

    if let Err(e) = resp.write_to(writer).await {
        error!("Error sending response to client: {e}");
    }
}

#[instrument]
async fn ipc_server(app: Arc<AppState>, sock: UnixListener) {
    loop {
        match sock.accept().await {
            Ok((conn, _)) => {
                tokio::spawn(handle_ipc_client(conn, app.clone()));
            }
            Err(e) => {
                error!("Couldn't accept socket connection: {}", e);
                continue;
            }
        }
    }
}

async fn quic_server(app: Arc<AppState>, mut notify: mpsc::Receiver<()>) {
    while notify.recv().await.is_some() {
        let mut active = app.active_config.write().await;
        let cfg = app.config.read().await.clone();
        *active = cfg;
        info!("Reloaded config!");
    }
}

#[derive(Debug, clap::Args)]
struct ConnectionAdd {
    #[arg(help = "The name of a registered remote")]
    remote: String,
    #[arg(short, long, help = "Specify a non-default port number")]
    port: u16,
}

#[derive(Subcommand, Debug)]
#[command(
    about = "Operations relating to tunnels",
    long_about = "Multiple port mappings can use the same tunnel. It is recommended to create separate tunnels for high traffic ports"
)]
enum Connection {
    Add(ConnectionAdd),
    List,
    Rm,
    Info,
}

#[derive(Subcommand, Debug)]
enum Commands {
    #[command(subcommand)]
    Connection(Connection),
    Remote,
}

#[derive(Parser, Debug)]
#[command(author = "Solomon W.", version, about = "The tunneling firewall")]
struct Cli {
    #[arg(long, help = "Run as the swim daemon")]
    daemonize: bool,
    #[arg(long, help = "Display the license")]
    license: bool,
    #[command(subcommand)]
    command: Option<Commands>,
}

#[tokio::main]
async fn main() -> Result<()> {
    let cli = Cli::parse();

    if cli.license {
        println!("{}", LICENSE);
        std::process::exit(0);
    }

    if cli.daemonize {
        let sock = UnixListener::bind(IPC_SOCKET_PATH).expect("Couldn't bind to unix socket!");
        let (tx, rx) = mpsc::channel(10);

        let config = Config::default();

        let app = Arc::new(AppState {
            config: Arc::new(RwLock::new(config)),
            notify: tx,
            active_config: Arc::new(RwLock::new(Config::default())),
        });

        // TODO: Early config init here

        let ipc_handle = tokio::spawn(ipc_server(app.clone(), sock));
        let quic_handle = tokio::spawn(quic_server(app.clone(), rx));

        ipc_handle.await.unwrap();
        quic_handle.await.unwrap();
    }

    if let Some(cmd) = cli.command {
        match cmd {
            Commands::Connection(_connection) => println!("Hello swim!"),
            _ => todo!("Not yet implemented!"),
        }
    }
    Ok(())
}
